"""create lswitch and lswitch port orphaned tables

Revision ID: 1284c81cf727
Revises: 4358d1b8cc75
Create Date: 2014-07-02 23:05:20.855269

"""

# revision identifiers, used by Alembic.
revision = '1284c81cf727'
down_revision = '4358d1b8cc75'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'quark_nvp_orphaned_lswitches',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('nvp_id', sa.String(length=36), nullable=False),
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('display_name', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index(
        'ix_quark_nvp_orphaned_lswitches_nvp_id',
        'quark_nvp_orphaned_lswitches',
        ['nvp_id'],
        unique=False)
    op.create_index(
        'ix_quark_nvp_orphaned_lswitches_network_id',
        'quark_nvp_orphaned_lswitches',
        ['network_id'],
        unique=False)
    op.create_index(
        'ix_quark_nvp_orphaned_lswitches_display_name',
        'quark_nvp_orphaned_lswitches',
        ['display_name'],
        unique=False)
    op.create_table(
        'quark_nvp_orphaned_lswitch_ports',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index(
        'ix_quark_nvp_orphaned_lswitch_ports_port_id',
        'quark_nvp_orphaned_lswitch_ports',
        ['port_id'],
        unique=False)


def downgrade():
    op.drop_index('ix_quark_nvp_orphaned_lswitches_nvp_id')
    op.drop_index('ix_quark_nvp_orphaned_lswitches_network_id')
    op.drop_index('ix_quark_nvp_orphaned_lswitches_display_name')
    op.drop_table('quark_nvp_orphaned_lswitches')
    op.drop_index('ix_quark_nvp_orphaned_lswitch_ports_port_id')
    op.drop_table('quark_nvp_orphaned_lswitch_ports')



# Copyright (c) 2013 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.api import extensions
from neutron.api.v2 import attributes as attrs
from neutron.extensions import securitygroup

RESOURCE_NAME = "port"
RESOURCE_COLLECTION = RESOURCE_NAME + "s"
EXTENDED_ATTRIBUTES_2_0 = {
    RESOURCE_COLLECTION: {
        "network_id": {"allow_post": True, "allow_put": False,
                       "default": '', "is_visible": True},
        "tenant_id": {"allow_post": True, "allow_put": False,
                      "default": '', "is_visible": True},
        "segment_id": {"allow_post": True, "allow_put": False,
                       "default": False},
        "bridge": {'allow_post': False, 'allow_put': False,
                   'default': False, 'is_visible': True},
        "device_owner": {'allow_post': True, 'allow_put': True,
                         'default': '', 'is_visible': True,
                         'enforce_policy': True},
        'use_forbidden_mac_range': {'allow_post': True, 'allow_put': True,
                                    'default': False,
                                    'convert_to': attrs.convert_to_boolean,
                                    'enforce_policy': True,
                                    'is_visible': True},
        "security_groups": {
            "allow_post": True,
            "allow_put": True,
            "is_visible": True,
            "enforce_policy": True,
            "convert_to": securitygroup.convert_to_uuid_list_or_none,
            "default": attrs.ATTR_NOT_SPECIFIED},
        "vlan_id": {
            "allow_post": False,
            "allow_put": False,
            "is_visible": True},
        "network_plugin": {"allow_post": True, "allow_put": False,
                           "enforce_policy": True,
                           "is_visible": False, "default": ''},
        "instance_node_id": {"allow_post": True, "allow_put": False,
                             "default": '', "is_visible": False},
    }
}


class Ports_quark(extensions.ExtensionDescriptor):
    """Extends ports for quark API purposes.

    """

    @classmethod
    def get_name(cls):
        return "Quark Ports API Extension"

    @classmethod
    def get_alias(cls):
        return "ports_quark"

    @classmethod
    def get_description(cls):
        return "Quark Ports API Extension"

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/network/ext/"
                "port_disassociate/api/v2.0")

    @classmethod
    def get_updated(cls):
        return "2013-03-25T19:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}

    @classmethod
    def get_request_extensions(cls):
        exts = []

        return exts



# Copyright 2015 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#

import mock

from quark.agent import agent
from quark.agent import xapi
from quark.tests import test_base


class TestAgentPartitionVifs(test_base.TestBase):
    @mock.patch("quark.agent.xapi.XapiClient._session")
    def test_partition_vifs(self, sess):
        def _vif_rec(mac, tagged):
            rec = {"MAC": mac, "other_config": {}}
            if tagged:
                rec["other_config"] = {"security_groups": "enabled"}
            return rec

        vif_recs = [_vif_rec(1, False), _vif_rec(2, True), _vif_rec(3, True),
                    _vif_rec(4, False), _vif_rec(5, False)]

        interfaces = [xapi.VIF("added", vif_recs[0], "added_ref"),
                      xapi.VIF("updated", vif_recs[1], "updated_ref"),
                      xapi.VIF("removed", vif_recs[2], "removed_ref"),
                      xapi.VIF("no groups", vif_recs[3], "no groups ref"),
                      xapi.VIF("self heal", vif_recs[4], "self heal ref")]

        sg_states = {interfaces[0]: False, interfaces[1]: False,
                     interfaces[4]: True}

        xapi_client = xapi.XapiClient()

        added, updated, removed = agent.partition_vifs(xapi_client,
                                                       interfaces,
                                                       sg_states)

        self.assertEqual(added, [interfaces[0], interfaces[4]])
        self.assertEqual(updated, [interfaces[1]])
        self.assertEqual(removed, [interfaces[2]])



"""Create enabled and do_not_use columns

Revision ID: 5927940a466e
Revises: 552b213c2b8c
Create Date: 2014-07-07 07:50:39.372294

"""

# revision identifiers, used by Alembic.
revision = '5927940a466e'
down_revision = '552b213c2b8c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('quark_port_ip_address_associations',
                  sa.Column('enabled',
                            sa.Boolean(),
                            nullable=False,
                            server_default='1'))
    op.add_column('quark_mac_address_ranges',
                  sa.Column('do_not_use',
                            sa.Boolean(),
                            nullable=False,
                            server_default='0'))


def downgrade():
    op.drop_column('quark_mac_address_ranges', 'do_not_use')
    op.drop_column('quark_port_ip_address_associations', 'enabled')



#!/usr/bin/python
# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Quark Redis Security Groups CLI tool.

Usage: redis_sg_tool [-h] [--config-file=PATH] [--retries=<retries>]
                     [--retry-delay=<delay>] <command> [--yarly]

Options:
    -h --help  Show this screen.
    --version  Show version.
    --config-file=PATH  Use a different config file path
    --retries=<retries>  Number of times to re-attempt some operations
    --retry-delay=<delay>  Amount of time to wait between retries

Available commands are:
    redis_sg_tool test-connection
    redis_sg_tool vifs-in-redis
    redis_sg_tool num-groups
    redis_sg_tool ports-with-groups
    redis_sg_tool purge-orphans [--yarly]
    redis_sg_tool write-groups [--yarly]
    redis_sg_tool -h | --help
    redis_sg_tool --version

"""
import sys
import time

import docopt
import netaddr
from neutron.common import config
import neutron.context
from oslo_config import cfg

from quark.cache import security_groups_client as sg_client
from quark.db import api as db_api
from quark import exceptions as q_exc

VERSION = 0.1
RETRIES = 5
RETRY_DELAY = 1


class QuarkRedisTool(object):
    def __init__(self, arguments):
        self._args = arguments

        self._retries = RETRIES
        self._retry_delay = RETRY_DELAY

        if self._args.get("--retries"):
            self._retries = int(self._args["--retries"])

        if self._args.get("--retry-delay"):
            self._retry_delay = int(self._args["--retry-delay"])

        config_args = []
        if self._args.get("--config-file"):
            config_args.append("--config-file=%s" %
                               self._args.pop("--config-file"))

        self._dryrun = not self._args.get("--yarly")

        config.init(config_args)
        if not cfg.CONF.config_file:
            sys.exit(_("ERROR: Unable to find configuration file via the "
                       "default search paths (~/.neutron/, ~/, /etc/neutron/, "
                       "/etc/) and the '--config-file' option!"))

    def dispatch(self):
        command = self._args.get("<command>")
        if command == "test-connection":
            self.test_connection()
        elif command == "vifs-in-redis":
            self.vif_count()
        elif command == "num-groups":
            self.num_groups()
        elif command == "ports-with-groups":
            self.ports_with_groups()
        elif command == "purge-orphans":
            self.purge_orphans(self._dryrun)
        elif command == "write-groups":
            self.write_groups(self._dryrun)
        else:
            print("Redis security groups tool. Re-run with -h/--help for "
                  "options")

    def _get_connection(self, giveup=True):
        client = sg_client.SecurityGroupsClient()
        try:
            if client.ping():
                return client
        except Exception as e:
            print(e)
            if giveup:
                print("Giving up...")
                sys.exit(1)

    def test_connection(self):
        client = self._get_connection()
        if client:
            print("Connected Successfully")
            return True
        else:
            print("Could not connect to Redis")
            return False

    def vif_count(self):
        client = self._get_connection()
        print(len(client.vif_keys(field=sg_client.SECURITY_GROUP_HASH_ATTR)))

    def num_groups(self):
        ctx = neutron.context.get_admin_context()
        print(db_api.security_group_count(ctx))

    def ports_with_groups(self):
        ctx = neutron.context.get_admin_context()
        print(db_api.ports_with_security_groups_count(ctx))

    def purge_orphans(self, dryrun=False):
        client = self._get_connection()
        ctx = neutron.context.get_admin_context()
        ports_with_groups = db_api.ports_with_security_groups_find(ctx).all()
        if dryrun:
            print()
            print("Purging orphans in dry run mode. Existing rules in Redis "
                  "will be checked against those in the database. If any "
                  "are found in Redis but lack matching database rules, "
                  "they'll be deleted from the database.\n\nTo actually "
                  "apply the groups, re-run with the --yarly flag.")
            print()
            print("Found %s ports with security groups" %
                  len(ports_with_groups))

        # Pre-spin the list of orphans
        vifs = {}
        for vif in client.vif_keys():
            vifs[vif] = False

        if dryrun:
            print("Found %d VIFs in Redis" % len(vifs))

        # Pop off the ones we find in the database
        for port in ports_with_groups:
            vif_key = client.vif_key(port["device_id"], port["mac_address"])
            vifs.pop(vif_key, None)

        if dryrun:
            print("Found %d orphaned VIF rule sets" % len(vifs))
            print('=' * 80)

        for orphan in vifs.keys():
            if dryrun:
                print("VIF %s is orphaned" % orphan)
            else:
                for retry in xrange(self._retries):
                    try:
                        client.delete_key(orphan)
                        break
                    except q_exc.RedisConnectionFailure:
                        time.sleep(self._retry_delay)
                        client = self._get_connection(giveup=False)
        if dryrun:
            print('=' * 80)
            print()
            print("Re-run with --yarly to apply changes")

        print("Done!")

    def write_groups(self, dryrun=False):
        client = self._get_connection()
        ctx = neutron.context.get_admin_context()
        ports_with_groups = db_api.ports_with_security_groups_find(ctx).all()
        if dryrun:
            print()
            print("Writing groups in dry run mode. Existing rules in Redis "
                  "will be checked against those in the database, with a "
                  "running report generated of all those that will be "
                  "overwritten.\n\nTo actually apply the groups, re-run "
                  "with the --yarly flag.")
            print()
            print("Found %s ports with security groups" %
                  len(ports_with_groups))

        if dryrun:
            vifs = len(client.vif_keys())
            if vifs > 0:
                print("There are %d VIFs with rules in Redis, some of which "
                      "may be overwritten!" % vifs)
                print()

        overwrite_count = 0
        for port in ports_with_groups:
            mac = netaddr.EUI(port["mac_address"])

            # Rather than loading everything in one giant chunk, we'll make
            # trips per port.
            group_ids = [g["id"] for g in port.security_groups]
            rules = db_api.security_group_rule_find(ctx, group_id=group_ids,
                                                    scope=db_api.ALL)

            if dryrun:
                existing_rules = client.get_rules_for_port(port["device_id"],
                                                           port["mac_address"])
                if existing_rules:
                    overwrite_count += 1
                    db_len = len(rules)
                    existing_len = len(existing_rules["rules"])
                    print("== Port ID:%s - MAC:%s - Device ID:%s - "
                          "Redis Rules:%d - DB Rules:%d" %
                          (port["id"], mac, port["device_id"], existing_len,
                           db_len))

            if not dryrun:
                for retry in xrange(self._retries):
                    try:
                        payload = client.serialize_rules(rules)
                        client.apply_rules(
                            port["device_id"], port["mac_address"], payload)
                        break
                    except q_exc.RedisConnectionFailure:
                        time.sleep(self._retry_delay)
                        client = self._get_connection(giveup=False)

        if dryrun:
            print()
            print("Total number of VIFs to overwrite/were overwritten: %s" %
                  overwrite_count)
            diff = vifs - overwrite_count
            if diff > 0:
                print("Orphaned VIFs in Redis:", diff)
                print("Run purge-orphans to clean then up")

        if dryrun:
            print("Total number of VIFs to write: %d" %
                  len(ports_with_groups))

        if dryrun:
            print('=' * 80)
            print("Re-run with --yarly to apply changes")
        print("Done!")


def main():
    arguments = docopt.docopt(__doc__,
                              version="Quark Redis CLI %.2f" % VERSION)
    redis_tool = QuarkRedisTool(arguments)
    redis_tool.dispatch()


if __name__ == "__main__":
    main()



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Provide strategies for allocating network segments. (vlan, vxlan, etc)
"""
from quark.db import api as db_api
from quark import exceptions as q_exc

from oslo_log import log as logging
from oslo_utils import timeutils

import itertools
import random

LOG = logging.getLogger(__name__)


class BaseSegmentAllocation(object):

    segment_type = None

    def _validate_range(self, context, sa_range):
        raise NotImplementedError()

    def _chunks(self, iterable, chunk_size):
        """Chunks data into chunk with size<=chunk_size."""
        iterator = iter(iterable)
        chunk = list(itertools.islice(iterator, 0, chunk_size))
        while chunk:
            yield chunk
            chunk = list(itertools.islice(iterator, 0, chunk_size))

    def _check_collisions(self, new_range, existing_ranges):
        """Check for overlapping ranges."""
        def _contains(num, r1):
            return (num >= r1[0] and
                    num <= r1[1])

        def _is_overlap(r1, r2):
            return (_contains(r1[0], r2) or
                    _contains(r1[1], r2) or
                    _contains(r2[0], r1) or
                    _contains(r2[1], r1))

        for existing_range in existing_ranges:
            if _is_overlap(new_range, existing_range):
                return True
        return False

    def _make_segment_allocation_dict(self, id, sa_range):
        return dict(
            id=id,
            segment_id=sa_range["segment_id"],
            segment_type=sa_range["segment_type"],
            segment_allocation_range_id=sa_range["id"],
            deallocated=True
        )

    def _populate_range(self, context, sa_range):
        first_id = sa_range["first_id"]
        last_id = sa_range["last_id"]
        id_range = xrange(first_id, last_id + 1)

        LOG.info("Starting segment allocation population for "
                 "range:%s size:%s."
                 % (sa_range["id"], len(id_range)))

        total_added = 0
        for chunk in self._chunks(id_range, 5000):
            sa_dicts = []
            for segment_id in chunk:
                sa_dict = self._make_segment_allocation_dict(
                    segment_id, sa_range)
                sa_dicts.append(sa_dict)
            db_api.segment_allocation_range_populate_bulk(context, sa_dicts)
            context.session.flush()
            total_added = total_added + len(sa_dicts)

            LOG.info("Populated %s/%s segment ids for range:%s"
                     % (total_added, len(id_range), sa_range["id"]))

        LOG.info("Finished segment allocation population for "
                 "range:%s size:%s."
                 % (sa_range["id"], len(id_range)))

    def _create_range(self, context, sa_range):
        with context.session.begin(subtransactions=True):
            # Validate any range-specific things, like min/max ids.
            self._validate_range(context, sa_range)

            # Check any existing ranges for this segment for collisions
            segment_id = sa_range["segment_id"]
            segment_type = sa_range["segment_type"]

            filters = {"segment_id": segment_id,
                       "segment_type": segment_type}
            existing_ranges = db_api.segment_allocation_range_find(
                context, lock_mode=True, scope=db_api.ALL, **filters)

            collides = self._check_collisions(
                (sa_range["first_id"], sa_range["last_id"]),
                [(r["first_id"], r["last_id"]) for r in existing_ranges])

            if collides:
                raise q_exc.InvalidSegmentAllocationRange(
                    msg=("The specified allocation collides with existing "
                         "range"))

            return db_api.segment_allocation_range_create(
                context, **sa_range)

    def create_range(self, context, sa_range):
        return self._create_range(context, sa_range)

    def populate_range(self, context, sa_range):
        return self._populate_range(context, sa_range)

    def _try_allocate(self, context, segment_id, network_id):
        """Find a deallocated network segment id and reallocate it.

        NOTE(morgabra) This locks the segment table, but only the rows
        in use by the segment, which is pretty handy if we ever have
        more than 1 segment or segment type.
        """
        LOG.info("Attempting to allocate segment for network %s "
                 "segment_id %s segment_type %s"
                 % (network_id, segment_id, self.segment_type))

        filter_dict = {
            "segment_id": segment_id,
            "segment_type": self.segment_type,
            "do_not_use": False
        }
        available_ranges = db_api.segment_allocation_range_find(
            context, scope=db_api.ALL, **filter_dict)
        available_range_ids = [r["id"] for r in available_ranges]

        try:
            with context.session.begin(subtransactions=True):
                # Search for any deallocated segment ids for the
                # given segment.
                filter_dict = {
                    "deallocated": True,
                    "segment_id": segment_id,
                    "segment_type": self.segment_type,
                    "segment_allocation_range_ids": available_range_ids
                }

                # NOTE(morgabra) We select 100 deallocated segment ids from
                # the table here, and then choose 1 randomly. This is to help
                # alleviate the case where an uncaught exception might leave
                # an allocation active on a remote service but we do not have
                # a record of it locally. If we *do* end up choosing a
                # conflicted id, the caller should simply allocate another one
                # and mark them all as reserved. If a single object has
                # multiple reservations on the same segment, they will not be
                # deallocated, and the operator must resolve the conficts
                # manually.
                allocations = db_api.segment_allocation_find(
                    context, lock_mode=True, **filter_dict).limit(100).all()

                if allocations:
                    allocation = random.choice(allocations)

                    # Allocate the chosen segment.
                    update_dict = {
                        "deallocated": False,
                        "deallocated_at": None,
                        "network_id": network_id
                    }
                    allocation = db_api.segment_allocation_update(
                        context, allocation, **update_dict)
                    LOG.info("Allocated segment %s for network %s "
                             "segment_id %s segment_type %s"
                             % (allocation["id"], network_id, segment_id,
                                self.segment_type))
                    return allocation
        except Exception:
            LOG.exception("Error in segment reallocation.")

        LOG.info("Cannot find reallocatable segment for network %s "
                 "segment_id %s segment_type %s"
                 % (network_id, segment_id, self.segment_type))

    def allocate(self, context, segment_id, network_id):
        allocation = self._try_allocate(
            context, segment_id, network_id)

        if allocation:
            return allocation

        raise q_exc.SegmentAllocationFailure(
            segment_id=segment_id, segment_type=self.segment_type)

    def _try_deallocate(self, context, segment_id, network_id):
        LOG.info("Attempting to deallocate segment for network %s "
                 "segment_id %s segment_type %s"
                 % (network_id, segment_id, self.segment_type))

        with context.session.begin(subtransactions=True):
            filter_dict = {
                "deallocated": False,
                "segment_id": segment_id,
                "segment_type": self.segment_type,
                "network_id": network_id
            }

            allocations = db_api.segment_allocation_find(
                context, **filter_dict).all()

            if not allocations:
                LOG.info("Could not find allocated segment for network %s "
                         "segment_id %s segment_type %s for deallocate."
                         % (network_id, segment_id, self.segment_type))
                return

            if len(allocations) > 1:
                LOG.error("Found multiple allocated segments for network %s "
                          "segment_id %s segment_type %s for deallocate. "
                          "Refusing to deallocate, these allocations are now "
                          "orphaned."
                          % (network_id, segment_id, self.segment_type))
                return

            allocation = allocations[0]
            # Deallocate the found segment.
            update_dict = {
                "deallocated": True,
                "deallocated_at": timeutils.utcnow(),
                "network_id": None
            }
            allocation = db_api.segment_allocation_update(
                context, allocation, **update_dict)

            LOG.info("Deallocated %s allocated segment(s) for network %s "
                     "segment_id %s segment_type %s"
                     % (len(allocations), network_id, segment_id,
                        self.segment_type))

    def deallocate(self, context, segment_id, network_id):
        self._try_deallocate(context, segment_id, network_id)


class VXLANSegmentAllocation(BaseSegmentAllocation):

    VXLAN_MIN = 1
    VXLAN_MAX = (2 ** 24) - 1

    segment_type = 'vxlan'

    def _validate_range(self, context, sa_range):
        # Validate that the range is legal and makes sense.
        try:
            first_id = sa_range["first_id"]
            last_id = sa_range["last_id"]
            first_id, last_id = (int(first_id), int(last_id))
            assert first_id >= self.VXLAN_MIN
            assert last_id <= self.VXLAN_MAX
            assert first_id <= last_id
        except Exception:
            raise q_exc.InvalidSegmentAllocationRange(
                msg="The specified allocation range is invalid")


class SegmentAllocationRegistry(object):
    def __init__(self):
        self.strategies = {
            VXLANSegmentAllocation.segment_type: VXLANSegmentAllocation(),
        }

    def is_valid_strategy(self, strategy_name):
        if strategy_name in self.strategies:
            return True
        return False

    def get_strategy(self, strategy_name):
        if self.is_valid_strategy(strategy_name):
            return self.strategies[strategy_name]
        raise Exception("Segment allocation strategy %s not found."
                        % (strategy_name))

REGISTRY = SegmentAllocationRegistry()



"""Create new port indexes

Revision ID: 41837dc547ce3
Revises: 3f0c11478a5d
Create Date: 2016-05-02 00:00:00

"""

# revision identifiers, used by Alembic.
revision = '41837dc547ce3'
down_revision = '3f0c11478a5d'

from alembic import op


def upgrade():
    op.create_index(op.f('ix_quark_ports_network_id_device_id'),
                    'quark_ports', ['network_id', 'device_id'])
    op.create_index(op.f('ix_quark_ports_device_id'),
                    'quark_ports', ['device_id'])


def downgrade():
    op.drop_index(op.f('ix_quark_ports_device_id'), table_name='quark_ports')
    op.drop_index(op.f('ix_quark_ports_network_id_device_id'),
                  table_name='quark_ports')



# Copyright (c) 2013 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.api import extensions
from neutron import manager
from neutron import wsgi
from oslo_log import log as logging
import quark.utils as utils
import webob

RESOURCE_NAME = 'ip_address'
RESOURCE_COLLECTION = RESOURCE_NAME + "es"
EXTENDED_ATTRIBUTES_2_0 = {
    RESOURCE_COLLECTION: {}
}

attr_dict = EXTENDED_ATTRIBUTES_2_0[RESOURCE_COLLECTION]
attr_dict[RESOURCE_NAME] = {'allow_post': True,
                            'allow_put': True,
                            'is_visible': True}

SUB_RESOURCE_ATTRIBUTE_MAP = {
    'ports': {
        'parent': {'collection_name': 'ip_addresses',
                   'member_name': 'ip_address'}
    }
}

LOG = logging.getLogger(__name__)


class IpAddressesController(wsgi.Controller):

    def __init__(self, plugin):
        self._resource_name = RESOURCE_NAME
        self._plugin = plugin

    @utils.exc_wrapper
    def index(self, request):
        context = request.context
        return {"ip_addresses":
                self._plugin.get_ip_addresses(context, **request.GET)}

    @utils.exc_wrapper
    def show(self, request, id):
        context = request.context
        return {"ip_address": self._plugin.get_ip_address(context, id)}

    @utils.exc_wrapper
    def create(self, request, body=None):
        body = self._deserialize(request.body, request.get_content_type())
        return {"ip_address": self._plugin.create_ip_address(
                request.context, body)}

    @utils.exc_wrapper
    def update(self, request, id, body=None):
        body = self._deserialize(request.body, request.get_content_type())
        return {"ip_address": self._plugin.update_ip_address(
                request.context, id, body)}

    @utils.exc_wrapper
    def delete(self, request, id):
        context = request.context
        return self._plugin.delete_ip_address(context, id)


class IpAddressPortController(wsgi.Controller):

    def __init__(self, plugin):
        self._resource_name = RESOURCE_NAME
        self._plugin = plugin

    def _clean_query_string(self, request, filters):
        clean_list = ['id', 'device_id', 'service']
        for clean in clean_list:
            if clean in request.GET:
                filters[clean] = request.GET[clean]
                del request.GET[clean]

    @utils.exc_wrapper
    def index(self, ip_address_id, request):
        context = request.context
        filters = {}
        self._clean_query_string(request, filters)
        fx = self._plugin.get_ports_for_ip_address
        ports = fx(context, ip_address_id, filters=filters, **request.GET)
        return {"ip_addresses_ports": ports}

    def create(self, request, **kwargs):
        raise webob.exc.HTTPNotImplemented()

    @utils.exc_wrapper
    def show(self, ip_address_id, request, id):
        context = request.context
        # TODO(jlh): need to ensure ip_address_id is used to filter port
        return {"ip_addresses_port":
                self._plugin.get_port_for_ip_address(context,
                                                     ip_address_id, id)}

    @utils.exc_wrapper
    def update(self, ip_address_id, request, id, body=None):
        body = self._deserialize(request.body, request.get_content_type())
        return {"ip_addresses_port": self._plugin.update_port_for_ip(
            request.context, ip_address_id, id, body)}

    def delete(self, request, id, **kwargs):
        raise webob.exc.HTTPNotImplemented()


class Ip_addresses(extensions.ExtensionDescriptor):
    """IP Addresses support."""
    @classmethod
    def get_name(cls):
        return "IP Addresses for a tenant"

    @classmethod
    def get_alias(cls):
        return RESOURCE_COLLECTION

    @classmethod
    def get_description(cls):
        return "Expose functions for tenant IP Address management"

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/network/ext/"
                "ip_addresses/api/v2.0")

    @classmethod
    def get_updated(cls):
        return "2013-02-19T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        ip_controller = IpAddressesController(
            manager.NeutronManager.get_plugin())
        ip_port_controller = IpAddressPortController(
            manager.NeutronManager.get_plugin())
        resources = []
        resources.append(extensions.ResourceExtension(
                         Ip_addresses.get_alias(),
                         ip_controller))
        parent = {'collection_name': 'ip_addresses',
                  'member_name': 'ip_address'}
        resources.append(extensions.ResourceExtension(
                         'ports', ip_port_controller, parent=parent))
        return resources



# Copyright (c) 2015 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.api import extensions
from neutron import manager
from neutron import wsgi
from neutron_lib import exceptions as n_exc
from oslo_log import log as logging

RESOURCE_NAME = "ip_availability"
RESOURCE_COLLECTION = "ip_availabilities"
EXTENDED_ATTRIBUTES_2_0 = {
    RESOURCE_COLLECTION: {}
}

attr_dict = EXTENDED_ATTRIBUTES_2_0[RESOURCE_COLLECTION]
attr_dict[RESOURCE_NAME] = {'allow_post': False,
                            'allow_put': False,
                            'is_visible': True}

LOG = logging.getLogger(__name__)


class IPAvailabilityController(wsgi.Controller):
    def __init__(self, plugin):
        self._resource_name = RESOURCE_NAME
        self._plugin = plugin

    def index(self, request):
        context = request.context
        if not context.is_admin:
            raise n_exc.NotAuthorized()
        return self._plugin.get_ip_availability(**request.GET)


class Ip_availability(extensions.ExtensionDescriptor):
    """IP Availability support."""
    @classmethod
    def get_name(cls):
        return "IP Availability for a Neutron deployment"

    @classmethod
    def get_alias(cls):
        return RESOURCE_COLLECTION

    @classmethod
    def get_description(cls):
        return ("Expose functions for cloud admin to get IP availability")

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/network/ext/"
                "ip-availability/api/v2.0")

    @classmethod
    def get_updated(cls):
        return "2015-02-18T00:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        plugin = manager.NeutronManager.get_plugin()
        controller = IPAvailabilityController(plugin)
        return [extensions.ResourceExtension(Ip_availability.get_alias(),
                                             controller)]



"""Ensure default IP policy for subnets with IP policies

Revision ID: 45a07fac3d38
Revises: 2748e48cee3a
Create Date: 2014-07-05 12:49:51.815631

"""

# revision identifiers, used by Alembic.
revision = '45a07fac3d38'
down_revision = '2748e48cee3a'

import logging

from alembic import op
from sqlalchemy.sql import column, select, table
import netaddr
from oslo_utils import timeutils
from oslo_utils import uuidutils
import sqlalchemy as sa

LOG = logging.getLogger("alembic.migration")


def upgrade():
    ip_policy = table('quark_ip_policy',
                      column('id', sa.String(length=36)),
                      column('tenant_id', sa.String(length=255)),
                      column('created_at', sa.DateTime()))
    ip_policy_cidrs = table('quark_ip_policy_cidrs',
                            column('id', sa.String(length=36)),
                            column('created_at', sa.DateTime()),
                            column('ip_policy_id', sa.String(length=36)),
                            column('cidr', sa.String(length=64)))
    subnets = table('quark_subnets',
                    column('_cidr', sa.String(length=64)),
                    column('ip_policy_id', sa.String(length=36)))

    connection = op.get_bind()

    # 1. Get all ip_policy_cidrs for subnets with an ip_policy.
    j = subnets.outerjoin(
        ip_policy_cidrs,
        subnets.c.ip_policy_id == ip_policy_cidrs.c.ip_policy_id)
    q = select([subnets.c.ip_policy_id, subnets.c._cidr,
                ip_policy_cidrs.c.id, ip_policy_cidrs.c.cidr]).select_from(
                    j).where(subnets.c.ip_policy_id != None).order_by(  # noqa
                        subnets.c.ip_policy_id)
    data = connection.execute(q).fetchall()
    if data is None:
        return

    # 2. Check ip_policy_cidrs contains default ip policy for subnet.
    ipp_to_update = dict()

    def _test_change_needed(ipp_id, s, ipp):
        if s is None or ipp is None:
            return
        updated = False
        last = netaddr.IPAddress(subnet.broadcast)
        first = netaddr.IPAddress(subnet.network)
        if last not in ipp:
            updated = True
            ipp.add(last)
        if first not in ipp:
            updated = True
            ipp.add(first)
        if updated:
            ipp_to_update[ipp_id] = ipp

    prev_ip_policy_id = ''
    subnet, ip_policy = None, None
    for ip_policy_id, cidr, ippc_id, ippc_cidr in data:
        if ip_policy_id != prev_ip_policy_id:
            _test_change_needed(prev_ip_policy_id, subnet, ip_policy)
            subnet, ip_policy = netaddr.IPNetwork(cidr), netaddr.IPSet()
        ip_policy |= netaddr.IPSet([ippc_cidr] if ippc_cidr else [])
        prev_ip_policy_id = ip_policy_id
    _test_change_needed(prev_ip_policy_id, subnet, ip_policy)

    if not ipp_to_update.keys():
        return

    LOG.info("IP Policy IDs to update: %s", ipp_to_update.keys())

    # 3. Delete ip_policy_cidrs for ip_policy_ids to be updated.
    connection.execute(ip_policy_cidrs.delete().where(
        ip_policy_cidrs.c.ip_policy_id.in_(ipp_to_update.keys())))

    # 4. Insert ip_policy_cidrs for ip_policy_ids to be updated.
    vals = [dict(id=uuidutils.generate_uuid(),
                 created_at=timeutils.utcnow(),
                 ip_policy_id=key,
                 cidr=str(x.cidr))
            for key in ipp_to_update.keys()
            for x in ipp_to_update[key].iter_cidrs()]
    if not vals:
        return

    LOG.info("IP Policy CIDR IDs to insert: %s", [v["id"] for v in vals])

    connection.execute(ip_policy_cidrs.insert(), *vals)


def downgrade():
    raise NotImplementedError()



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from quark.drivers import base
from quark.tests import test_base


class TestBaseDriver(test_base.TestBase):
    def setUp(self):
        super(TestBaseDriver, self).setUp()
        self.driver = base.BaseDriver()

    def test_load_config(self):
        self.driver.load_config()

    def test_get_connection(self):
        self.driver.get_connection()

    def test_select_ipam_strategy(self):
        strategy = self.driver.select_ipam_strategy(1, "ANY")
        self.assertEqual(strategy, "ANY")

    def test_create_network(self):
        self.driver.create_network(context=self.context, network_name="public")

    def test_delete_network(self):
        self.driver.delete_network(context=self.context, network_id=1)

    def test_create_port(self):
        self.driver.create_port(context=self.context, network_id=1, port_id=2)

    def test_update_port(self):
        self.driver.update_port(context=self.context, network_id=1, port_id=2)

    def test_delete_port(self):
        self.driver.delete_port(context=self.context, port_id=2)

    def test_diag_network(self):
        diag = self.driver.diag_network(self.context, network_id=1)
        self.assertEqual(diag, {})

    def test_diag_port(self):
        diag = self.driver.diag_port(self.context, network_id=1)
        self.assertEqual(diag, {})

    def test_create_security_group(self):
        self.driver.create_security_group(context=self.context,
                                          group_name="mygroup")

    def test_delete_security_group(self):
        self.driver.delete_security_group(context=self.context,
                                          group_id=3)

    def test_update_security_group(self):
        self.driver.update_security_group(context=self.context,
                                          group_id=3)

    def test_create_security_group_rule(self):
        rule = {'ethertype': 'IPv4', 'direction': 'ingress'}
        self.driver.create_security_group_rule(context=self.context,
                                               group_id=3,
                                               rule=rule)

    def test_delete_security_group_rule(self):
        rule = {'ethertype': 'IPv4', 'direction': 'ingress'}
        self.driver.delete_security_group_rule(context=self.context,
                                               group_id=3,
                                               rule=rule)



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple
import contextlib

from oslo_config import cfg
from oslo_log import log as logging
import XenAPI


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

agent_opts = [
    cfg.StrOpt("xapi_connection_url"),
    cfg.StrOpt("xapi_connection_username", default="root"),
    cfg.IntOpt("xapi_enable_groups_retries", default=5),
    cfg.StrOpt("xapi_connection_password")
]

CONF.register_opts(agent_opts, "AGENT")
SECURITY_GROUPS_KEY = "security_groups"
VM = namedtuple('VM', ['ref', 'uuid', 'vifs', 'dom_id'])


class VIF(object):
    SEPARATOR = "."

    def __init__(self, device_id, record, ref):
        """Constructs VIF

        `device_id` and `mac_address` should be strings if they will later be
        compared to decoded VIF instances (via from_string).

        `ref` is the OpaqueRef string for the vif as returned from xenapi.
        """

        self.device_id = device_id
        self.record = record
        self.ref = ref
        self.success = False

    def __str__(self):
        return "%s%s%s%s%s" % (self.device_id, self.SEPARATOR,
                               self.mac_address, self.SEPARATOR,
                               self.ref)

    @property
    def mac_address(self):
        return self.record["MAC"]

    @property
    def tagged(self):
        return self.record["other_config"].get(SECURITY_GROUPS_KEY)

    @classmethod
    def from_string(cls, s):
        device_id, mac_address, ref = s.split(cls.SEPARATOR)
        return cls(device_id, mac_address, ref)

    def succeed(self):
        self.success = True

    def __repr__(self):
        return "VIF(%r, %r, %r)" % (self.device_id, self.mac_address,
                                    self.ref)

    def __eq__(self, other):
        return (self.device_id == other.device_id and
                self.mac_address == other.mac_address)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.device_id, self.mac_address))


class XapiClient(object):
    SECURITY_GROUPS_VALUE = "enabled"

    def __init__(self):
        with self.sessioned() as session:
            self._host_ref = session.xenapi.session.get_this_host(
                session.handle)
            self._host_uuid = session.xenapi.host.get_uuid(self._host_ref)

    def _session(self):
        LOG.debug("Created new Xapi session")

        session = XenAPI.Session(CONF.AGENT.xapi_connection_url)
        session.login_with_password(CONF.AGENT.xapi_connection_username,
                                    CONF.AGENT.xapi_connection_password)
        return session

    @contextlib.contextmanager
    def sessioned(self):
        session = None
        try:
            session = self._session()
            yield session
        except Exception:
            LOG.exception("Failed to create a XAPI session")
            raise
        finally:
            if session is not None:
                session.xenapi.session.logout()

    def get_instances(self, session):
        """Returns a dict of `VM OpaqueRef` (str) -> `xapi.VM`."""
        LOG.debug("Getting instances from Xapi")

        recs = session.xenapi.VM.get_all_records()

        # NOTE(asadoughi): Copied from xen-networking-scripts/utils.py
        is_inst = lambda r: (r['power_state'].lower() == 'running' and
                             not r['is_a_template'] and
                             not r['is_control_domain'] and
                             ('nova_uuid' in r['other_config'] or
                              r['name_label'].startswith('instance-')))
        instances = dict()
        for vm_ref, rec in recs.iteritems():
            if not is_inst(rec):
                continue
            instances[vm_ref] = VM(ref=vm_ref,
                                   uuid=rec["other_config"]["nova_uuid"],
                                   vifs=rec["VIFs"],
                                   dom_id=rec["domid"])
        return instances

    def get_interfaces(self):
        """Returns a set of VIFs from `get_instances` return value."""
        LOG.debug("Getting interfaces from Xapi")

        with self.sessioned() as session:
            instances = self.get_instances(session)
            recs = session.xenapi.VIF.get_all_records()

        interfaces = set()
        for vif_ref, rec in recs.iteritems():
            vm = instances.get(rec["VM"])
            if not vm:
                continue
            device_id = vm.uuid
            interfaces.add(VIF(device_id, rec, vif_ref))
        return interfaces

    def _set_security_groups(self, session, interfaces):
        LOG.debug("Setting security groups on %s", interfaces)

        for vif in interfaces:
            try:
                session.xenapi.VIF.add_to_other_config(
                    vif.ref, SECURITY_GROUPS_KEY,
                    self.SECURITY_GROUPS_VALUE)
            except XenAPI.Failure:
                # We shouldn't lose all of them because one failed
                # An example of a continuable failure is the VIF was deleted
                # in the (albeit very small) window between the initial fetch
                # and here.
                LOG.exception("Failed to enable security groups for VIF "
                              "with MAC %s" % vif.mac_address)

    def _unset_security_groups(self, session, interfaces):
        LOG.debug("Unsetting security groups on %s", interfaces)

        for vif in interfaces:
            try:
                session.xenapi.VIF.remove_from_other_config(
                    vif.ref,
                    SECURITY_GROUPS_KEY)
            except XenAPI.Failure:
                # NOTE(mdietz): RM11399 - removing a parameter that doesn't
                #               exist is idempotent. Trying to remove it
                #               from a VIF that doesn't exist raises :-( This
                #               may be a consequence of bad data, but we should
                #               try to cover ourselves here and continue.
                LOG.exception("Failed to disable security groups for VIF "
                              "with MAC %s" % vif.mac_address)

    def _refresh_interfaces(self, session, interfaces):
        LOG.debug("Refreshing devices on %s", interfaces)

        for vif in interfaces:
            try:
                vif_rec = session.xenapi.VIF.get_record(vif.ref)
                vm_rec = session.xenapi.VM.get_record(vif_rec["VM"])
                vif_index = vif_rec["device"]
                dom_id = vm_rec["domid"]
                vif.succeed()
            except XenAPI.Failure:
                LOG.exception("Failure when looking up VMs or VIFs")
                continue

            args = {"dom_id": dom_id, "vif_index": vif_index}
            session.xenapi.host.call_plugin(
                self._host_ref,
                "neutron_vif_flow",
                "online_instance_flows",
                args)

    def update_interfaces(self, added_sg, updated_sg, removed_sg):
        """Handles changes to interfaces' security groups

        Calls refresh_interfaces on argument VIFs. Set security groups on
        added_sg's VIFs. Unsets security groups on removed_sg's VIFs.
        """
        if not (added_sg or updated_sg or removed_sg):
            return

        with self.sessioned() as session:
            self._set_security_groups(session, added_sg)
            self._unset_security_groups(session, removed_sg)
            combined = added_sg + updated_sg + removed_sg
            self._refresh_interfaces(session, combined)
